<?php

/**
 * @file
 * Contains \Drupal\yamlform\Form\YamlFormSubmissionEmailForm.
 */

namespace Drupal\yamlform\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\yamlform\YamlFormSubmissionInterface;

/**
 * Defines a form that email's a YAML form submission.
 */
class YamlFormSubmissionEmailForm extends FormBase {

  /**
   * A YAML form submission.
   *
   * @var \Drupal\yamlform\YamlFormSubmissionInterface
   */
  protected $yamlformSubmission;

  /**
   * Associative array of emails being sent by the YAML form submission.
   *
   * @var array
   */
  protected $emails;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'yamlform_submission_email';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, YamlFormSubmissionInterface $yamlform_submission = NULL) {
    $this->yamlformSubmission = $yamlform_submission;
    /** @var \Drupal\yamlform\Plugin\YamlFormHandler\EmailYamlFormHandler[] $email_handlers */
    $email_handlers = $yamlform_submission->getYamlForm()->getHandlers('email');

    $this->emails = [];
    foreach ($email_handlers as $index => $email_handler) {
      $this->emails[$index] = $email_handler->getEmail($yamlform_submission);
      $this->emails[$index]['title'] = $email_handler->label();
      $this->emails[$index]['description'] = $email_handler->description();
    }

    // Get the (default) first email key/index.
    reset($this->emails);
    $mail_key = key($this->emails);

    $email_options = [];
    foreach ($this->emails as $index => $email) {
      $t_args = [
        '@title' => $email['title'],
        '@description' => ($email['description']) ? ': ' . $email['description'] : '',
        '@subject' => $email['subject'],
        '@to' => $email['to_mail'],
        '@from' => $email['from_mail'],
      ];
      $build = [
        'title' => [
          '#markup' => $this->t('<b>@title</b>@description', $t_args),
          '#suffix' => '<br/>',
        ],
        'mail' => [
          '#markup' => $this->t('@subject (To: @to)', $t_args),
        ],
      ];
      $email_options[$index] = \Drupal::service('renderer')->render($build);
    }
    $form['mail_key'] = [
      '#type' => 'radios',
      '#title' => 'Select an email message',
      '#options' => $email_options,
      '#default_value' => $mail_key,
      '#ajax' => [
        'callback' => '::updateMail',
        'wrapper' => 'edit-yamlform-mail-wrapper',
      ],
    ];

    $mail = $this->emails[$mail_key];
    $form['mail'] = [
      '#type' => 'details',
      '#title' => 'Email',
      '#open' => TRUE,
      '#tree' => TRUE,
      '#prefix' => '<div id="edit-yamlform-mail-wrapper">',
      '#suffix' => '</div>',
    ];
    $form['mail']['subject'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Subject'),
      '#default_value' => $mail['subject'],
    ];
    $form['mail']['to_mail'] = [
      '#type' => 'email',
      '#title' => $this->t('To email'),
      '#default_value' => $mail['to_mail'],
    ];
    $form['mail']['from_mail'] = [
      '#type' => 'email',
      '#title' => $this->t('From email '),
      '#required' => TRUE,
      '#default_value' => $mail['from_mail'],
    ];
    $form['mail']['from_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('From name'),
      '#required' => TRUE,
      '#default_value' => $mail['from_name'],
    ];
    $form['mail']['body'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Message'),
      '#rows' => 20,
      '#required' => TRUE,
      '#default_value' => $mail['body'],
    ];
    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Resend email'),
    ];

    // Add submission navigation.
    $form['navigation'] = [
      '#theme' => 'yamlform_submission_navigation',
      '#yamlform_submission' => $yamlform_submission,
      '#rel' => 'email-form',
      '#weight' => -20,
    ];
    $form['#attached']['library'][] = 'yamlform/yamlform';

    return $form;
  }

  /**
   * Handles switching between emails.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function updateMail(array $form, FormStateInterface $form_state) {
    $mail_key = $form_state->getValue('mail_key');
    foreach ($this->emails[$mail_key] as $key => $value) {
      if (isset($form['mail'][$key])) {
        $form['mail'][$key]['#value'] = $value;
      }
    }
    return $form['mail'];
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $mail_key = $form_state->getValue('mail_key');
    $params = $form_state->getValue('mail');

    $email_handlers = $this->yamlformSubmission->getYamlForm()->getHandlers('email');
    /** @var \Drupal\yamlform\Plugin\YamlFormHandler\EmailYamlFormHandler $email_handler */
    $email_handler = $email_handlers->get($mail_key);

    $email_handler->mail($params);

    $t_args = [
      '%to_mail' => $params['to_mail'],
      '%subject' => $params['subject'],
    ];
    drupal_set_message($this->t('Successfully re-sent %subject to %to_mail.', $t_args));
  }

}
