<?php

/**
 * @file
 * Contains \Drupal\yamlform\YamlFormSubmissionExporter.
 */

namespace Drupal\yamlform;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\Query\QueryFactory;
use Drupal\Core\Entity\EntityManagerInterface;
use Drupal\yamlform\Entity\YamlFormSubmission;
use Drupal\yamlform\Utility\YamlFormEntityReferenceHelper;
use Drupal\yamlform\Utility\YamlFormOptionsHelper;

/**
 * Exporter for YAML form submission results.
 */
class YamlFormSubmissionExporter {

  /**
   * The configuration object factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * YAML form submission storage.
   *
   * @var \Drupal\yamlform\YamlFormSubmissionStorage
   */
  protected $entityStorage;

  /**
   * The entity query factory.
   *
   * @var \Drupal\Core\Entity\Query\QueryFactory
   */
  protected $queryFactory;

  /**
   * YAML form element manager.
   *
   * @var \Drupal\yamlform\YamlFormElementManager
   */
  protected $yamlformElementManager;

  /**
   * Constructs a YamlFormSubmissionExporter object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration object factory.
   * @param \Drupal\Core\Entity\EntityManagerInterface $entity_manager
   *   The entity manager.
   * @param \Drupal\Core\Entity\Query\QueryFactory $query_factory
   *   The entity query factory.
   * @param \Drupal\yamlform\YamlFormElementManager $yamlform_element_manager
   *   The YAML form element manager.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityManagerInterface $entity_manager, QueryFactory $query_factory, YamlFormElementManager $yamlform_element_manager) {
    $this->configFactory = $config_factory;
    $this->entityStorage = $entity_manager->getStorage('yamlform_submission')->getFieldDefinitions();
    $this->queryFactory = $query_factory;
    $this->yamlformElementManager = $yamlform_element_manager;
  }

  /**
   * Get the number of submissions to be exported with each batch.
   *
   * @return int
   *   Number of submissions to be exported with each batch.
   */
  public function getBatchLimit() {
    return $this->configFactory->get('yamlform.settings')->get('batch.export_limit') ?: 500;
  }

  /**
   * Get default options for exporting a CSV.
   *
   * @return array
   *   Default options for exporting a CSV.
   */
  public function getDefaultExportOptions() {
    return [
      'delimiter' => ',',
      'header_keys' => 'label',
      'options_format' => 'compact',
      'options_item_format' => 'label',
      'entity_reference_format' => 'link',
      'excluded_columns' => [
        'uuid' => 'uuid',
        'token' => 'token',
        'changed' => 'changed',
        'data' => 'data',
        'yamlform_id' => 'yamlform_id',
      ],
      'range_type' => 'all',
      'range_latest' => 0,
      'range_start' => '',
      'range_end' => '',
      'state' => 'all',
      'download' => TRUE,
    ];
  }

  /**
   * Determine if YAML form submissions must be exported using batch processing.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   *
   * @return bool
   *   TRUE if YAML form submissions must be exported using batch processing.
   */
  public function requiresBatch(YamlFormInterface $yamlform) {
    return ($this->getTotal($yamlform, $this->getDefaultExportOptions()) > $this->getBatchLimit()) ? TRUE : FALSE;
  }

  /**
   * Get CSV file name and path for a YAML form.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   *
   * @return string
   *   CSV file name and path for a YAML form
   */
  public function getFilePath(YamlFormInterface $yamlform) {
    return file_directory_temp() . '/' . $yamlform->id() . '.csv';
  }

  /**
   * Generate YAML form submission as a CSV and write it to a temp file.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   * @param array $export_options
   *   An associative array of export options generated via the
   *   Drupal\yamlform\Form\YamlFormResultsExportForm.
   */
  public function generate(YamlFormInterface $yamlform, array $export_options) {
    $field_definitions = $this->getFieldDefinitions($export_options);
    $input_columns = $this->getInputColumns($yamlform, $export_options);

    // Build header and add to CSV.
    $this->writeHeader($yamlform, $field_definitions, $input_columns, $export_options);

    // Build records and add to CSV.
    $entity_ids = $this->getQuery($yamlform, $export_options)->execute();
    $yamlform_submissions = YamlFormSubmission::loadMultiple($entity_ids);
    $this->writeRecords($yamlform, $yamlform_submissions, $field_definitions, $input_columns, $export_options);
  }

  /**
   * Total number of submissions to be exported.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return int
   *   The total number of submissions to be exported.
   */
  public function getTotal(YamlFormInterface $yamlform, array $export_options) {
    return $this->getQuery($yamlform, $export_options)->count()->execute();
  }

  /**
   * Write YAML form results header to CSV file.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   The YAML form.
   * @param array $field_definitions
   *   YAML form submission field definitions.
   * @param array $input_columns
   *   YAML form inputs as columns.
   * @param array $export_options
   *   An associative array of export options.
   */
  public function writeHeader(YamlFormInterface $yamlform, array $field_definitions, array $input_columns, array $export_options) {
    $file_path = $this->getFilePath($yamlform);

    // Build header and add to CSV.
    $handle = fopen($file_path, 'w');
    $header = $this->buildHeader($field_definitions, $input_columns, $export_options);
    fputcsv($handle, $header, $export_options['delimiter']);
    fclose($handle);
  }

  /**
   * Write YAML form results header to CSV file.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   The YAML form.
   * @param array $yamlform_submissions
   *   An array containing YAML form submissions.
   * @param array $field_definitions
   *   YAML form submission field definitions.
   * @param array $input_columns
   *   YAML form inputs as columns.
   * @param array $export_options
   *   An associative array of export options.
   */
  public function writeRecords(YamlFormInterface $yamlform, array $yamlform_submissions, array $field_definitions, array $input_columns, array $export_options) {
    $file_path = $this->getFilePath($yamlform);
    $handle = fopen($file_path, 'a');
    foreach ($yamlform_submissions as $yamlform_submission) {
      $record = $this->buildRecord($yamlform_submission, $field_definitions, $input_columns, $export_options);
      fputcsv($handle, $record, $export_options['delimiter']);
    }
    fclose($handle);
  }

  /**
   * Get YAML form submission field definitions.
   *
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return array
   *   An associative array containing YAML form submission field defintions.
   */
  public function getFieldDefinitions(array $export_options) {
    $field_definitions = $this->entityStorage;
    $field_definitions = array_diff_key($field_definitions, $export_options['excluded_columns']);

    // Add custom entity reference field definitions which rely on the
    // entity type and entity id.
    if ($export_options['entity_reference_format'] == 'link' && isset($field_definitions['entity_type']) && isset($field_definitions['entity_id'])) {
      $field_definitions['entity_title'] = [
        'name' => 'entity_title',
        'title' => t('Submitted to: Entity title'),
        'type' => 'entity_title',
      ];
      $field_definitions['entity_url'] = [
        'name' => 'entity_url',
        'title' => t('Submitted to: Entity URL'),
        'type' => 'entity_url',
      ];
    }
    return $field_definitions;
  }

  /**
   * Get YAML form inputs as columns.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return array
   *   An associative array containing YAML form inputs as columns.
   */
  public function getInputColumns(YamlFormInterface $yamlform, array $export_options) {
    $input_columns = $yamlform->getInputColumns();
    return array_diff_key($input_columns, $export_options['excluded_columns']);
  }

  /**
   * Get YAML form submission query for specified YAMl form and export options.
   *
   * @param \Drupal\yamlform\YamlFormInterface $yamlform
   *   A YAML form.
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return \Drupal\Core\Entity\Query\QueryInterface
   *   A YAML form submission entity query.
   */
  public function getQuery(YamlFormInterface $yamlform, array $export_options) {
    $query = $this->queryFactory->get('yamlform_submission')->condition('yamlform_id', $yamlform->id());

    // Filter by sid or date range.
    switch ($export_options['range_type']) {
      case 'sid':
        if ($export_options['range_start']) {
          $query->condition('sid', $export_options['range_start'], '>=');
        }
        if ($export_options['range_end']) {
          $query->condition('sid', $export_options['range_end'], '<=');
        }
        break;

      case 'date':
        if ($export_options['range_start']) {
          $query->condition('created', strtotime($export_options['range_start']), '>=');
        }
        if ($export_options['range_end']) {
          $query->condition('created', strtotime($export_options['range_end']), '<=');
        }
        break;
    }

    // Filter by (completion) state.
    switch ($export_options['state']) {
      case 'draft';
        $query->condition('in_draft', 1);
        break;

      case 'completed';
        $query->condition('in_draft', 0);
        break;

    }

    // Filter by latest.
    if ($export_options['range_type'] == 'latest' && $export_options['range_latest']) {
      $query->range(0, $export_options['range_latest']);
      $query->sort('sid', 'DESC');
    }
    else {
      $query->sort('sid');
    }

    return $query;
  }

  /****************************************************************************/
  // Header functions.
  /****************************************************************************/

  /**
   * Build CSV header using YAML form submission field definitions and YAML form input columns.
   *
   * @param array $field_definitions
   *   An associative array containing YAML form submission field definitions.
   * @param array $input_columns
   *   An associative array containing YAML form input columns.
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return array
   *   An array containing the CSV header.
   */
  protected function buildHeader(array $field_definitions, array $input_columns, array $export_options) {
    $header = [];
    foreach ($field_definitions as $field_definition) {
      if ($field_definition['type'] == 'entity_reference') {
        $this->formatHeaderEntityReference($header, $field_definition['title'], $field_definition['name'], $export_options);
      }
      else {
        $header[] = ($export_options['header_keys'] == 'label') ? $field_definition['title'] : $field_definition['name'];
      }
    }

    // Build input columns headers.
    foreach ($input_columns as $input_column) {
      if ($export_options['options_format'] == 'separate' && isset($input_column['element']['#options'])) {
        $element_options = $input_column['element']['#options'];
        foreach ($element_options as $option_value => $option_text) {
          $header[] = ($export_options['options_item_format'] == 'key') ? $option_value : $option_text;
        }
      }
      elseif ($input_column['element']['#type'] == 'entity_autocomplete' && empty($input_column['element']['#tags'])) {
        $this->formatHeaderEntityReference($header, $input_column['title'], $input_column['name'], $export_options);
      }
      else {
        $header[] = ($export_options['header_keys'] == 'label') ? $input_column['title'] : $input_column['name'];
      }
    }
    return $header;
  }

  /**
   * Format header entity reference.
   *
   * @param array $header
   *   Export header.
   * @param string $title
   *   Column title.
   * @param string $name
   *   Column name.
   * @param array $export_options
   *   An associative array of export options.
   */
  protected function formatHeaderEntityReference(array &$header, $title, $name, array $export_options) {
    if ($export_options['entity_reference_format'] == 'link') {
      if ($export_options['header_keys'] == 'label') {
        $header[] = $title . ' ID';
        $header[] = $title . ' Title';
        $header[] = $title . ' URL';
      }
      else {
        $header[] = $name . '_id';
        $header[] = $name . '_title';
        $header[] = $name . '_url';
      }
    }
    else {
      $header[] = ($export_options['header_keys'] == 'label') ? $title : $name;
    }
  }

  /****************************************************************************/
  // Record functions.
  /****************************************************************************/

  /**
   * Build CSV record using YAML form submission, field definitions, and input columns.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   * @param array $field_definitions
   *   An associative array containing YAML form submission field definitions.
   * @param array $input_columns
   *   An associative array containing YAML form input columns.
   * @param array $export_options
   *   An associative array of export options.
   *
   * @return array
   *   An array containing the CSV record.
   */
  protected function buildRecord(YamlFormSubmissionInterface $yamlform_submission, array $field_definitions, array $input_columns, array $export_options) {
    $record = [];

    // Build record field definition columns.
    foreach ($field_definitions as $field_definition) {
      $this->formatRecordFieldDefinitionValue($record, $yamlform_submission, $field_definition, $export_options);
    }

    // Build record input columns.
    $data = $yamlform_submission->getData();
    foreach ($input_columns as $column_name => $input_column) {
      $value = (isset($data[$column_name])) ? $data[$column_name] : '';

      // Handle select, radios and checkbox options.
      if (isset($input_column['element']['#options'])) {
        $this->formatRecordOptions($record, $value, $input_column, $export_options);
      }
      elseif ($input_column['element']['#type'] == 'entity_autocomplete') {
        $this->formatRecordEntityAutocomplete($record, $value, $input_column, $export_options);
      }
      else {
        $value = $this->yamlformElementManager->invokeMethod('formatText', $input_column['element'], $value);
        $record[] = $value;
      }
    }
    return $record;
  }

  /**
   * Get the field definition value from a YAML form submission entity.
   *
   * @param array $record
   *   The record to be added to the CSV export.
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   * @param array $field_definition
   *   The field definition for the value.
   * @param array $export_options
   *   An associative array of export options.
   */
  protected function formatRecordFieldDefinitionValue(array &$record, YamlFormSubmissionInterface $yamlform_submission, array $field_definition, array $export_options) {
    $field_name = $field_definition['name'];
    $field_type = $field_definition['type'];
    switch ($field_type) {
      case 'created':
      case 'changed':
        $record[] = date('c', $yamlform_submission->get($field_name)->value);
        break;

      case 'entity_reference':
        $target_type = $field_definition['target_type'];
        $target_id = $yamlform_submission->get($field_name)->target_id;
        $default_value = "$target_type:$target_id";
        $this->formatRecordEntityReference($record, $target_type, $target_id, $default_value, $export_options);
        break;

      case 'entity_url':
      case 'entity_title':
        if (empty($yamlform_submission->entity_type->value) || empty($yamlform_submission->entity_id->value)) {
          $record[] = '';
          break;
        }

        $entity = entity_load($yamlform_submission->entity_type->value, $yamlform_submission->entity_id->value);
        if ($entity) {
          $record[] = ($field_type == 'entity_url') ? $entity->toUrl()->setOption('absolute', TRUE)->toString() : $entity->label();
        }
        else {
          $record[] = '';
        }
        break;

      default:
        $record[] = $yamlform_submission->get($field_name)->value;
        break;

    }
  }

  /**
   * Format exported submission's record options.
   *
   * @param array $record
   *   Exported submission's record.
   * @param mixed $value
   *   The submitted value.
   * @param array $column
   *   The column/element for the submitted value.
   * @param array $export_options
   *   An associative array of export options.
   */
  protected function formatRecordOptions(array &$record, $value, array $column, array $export_options) {
    $element_options = $column['element']['#options'];
    if ($export_options['options_format'] == 'separate') {
      // Separate multiple values (ie options).
      foreach ($element_options as $option_value => $option_text) {
        if (is_array($value) && isset($value[$option_value])) {
          $record[] = 'X';
        }
        elseif ($value == $option_value) {
          $record[] = 'X';
        }
        else {
          $record[] = '';
        }
      }
    }
    else {
      // Handle multiple values with options.
      if (is_array($value)) {
        if ($export_options['options_item_format'] == 'label') {
          $value = YamlFormOptionsHelper::getOptionsText($value, $element_options);
        }
        $record[] = implode(',', $value);
      }
      // Handle single values with options.
      elseif ($export_options['options_item_format'] == 'label') {
        $record[] = YamlFormOptionsHelper::getOptionText($value, $element_options);
      }
    }
  }

  /**
   * Format exported submission's record entity (reference) autocomplete.
   *
   * @param array $record
   *   Exported submission's record.
   * @param mixed $value
   *   The submitted value.
   * @param array $column
   *   The column/element for the submitted value.
   * @param array $export_options
   *   An associative array of export options.
   */
  protected function formatRecordEntityAutocomplete(array &$record, $value, array $column, array $export_options) {
    $target_type = $column['element']['#target_type'];
    if (!empty($column['element']['#tags'])) {
      $record[] = YamlFormEntityReferenceHelper::formatTargetEntities($target_type, $value, FALSE);
    }
    else {
      $target_id = $value;
      $default_value = "$target_type:$target_id";
      $this->formatRecordEntityReference($record, $target_type, $target_id, $default_value, $export_options);
    }
  }

  /**
   * Format entity reference as a record.
   *
   * @param array $record
   *   Exported submission's record.
   * @param string $entity_type
   *   The entity type.
   * @param string|int $entity_id
   *   The entity value.
   * @param string|int $default_value
   *   Default value to be displayed if entity is not available.
   * @param array $export_options
   *   An associative array of export options.
   */
  protected function formatRecordEntityReference(array &$record, $entity_type, $entity_id, $default_value, array $export_options) {
    if ($export_options['entity_reference_format'] == 'link') {
      if ($entity = entity_load($entity_type, $entity_id)) {
        $record[] = $entity->id();
        $record[] = $entity->label();
        $record[] = $entity->toUrl('canonical', ['absolute' => TRUE])->toString();
      }
      else {
        $record[] = $default_value;
        $record[] = '';
        $record[] = '';
      }
    }
    else {
      $record[] = $default_value;
    }
  }

}
