<?php

/**
 * @file
 * Contains \Drupal\yamlform\YamlFormSubmissionStorage.
 */

namespace Drupal\yamlform;

use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\Core\Session\AccountInterface;

/**
 * Defines a Controller class for YAML form submissions.
 */
class YamlFormSubmissionStorage extends SqlContentEntityStorage {

  /**
   * Get YAML form submission entity field definitions.
   *
   * The helper method is generally used for exporting results.
   *
   * See: \Drupal\yamlform\Element\YamlFormExcludedColumns
   * See: \Drupal\yamlform\Controller\YamlFormResultsExportController
   *
   * @return array
   *   An associative array of field definition key by field name containing
   *   title, name, and datatype.
   */
  public function getFieldDefinitions() {
    /** @var \Drupal\Core\Field\BaseFieldDefinition[] $definitions */
    $field_definitions = $this->entityManager->getBaseFieldDefinitions('yamlform_submission');

    // For now never let any see or export the serialize YAML data field.
    unset($field_definitions['data']);

    $definitions = [];
    foreach ($field_definitions as $field_name => $field_definition) {
      $definitions[$field_name] = [
        'title' => $field_definition->getLabel(),
        'name' => $field_name,
        'type' => $field_definition->getType(),
        'target_type' => $field_definition->getSetting('target_type'),
      ];
    }

    return $definitions;
  }

  /**
   * Delete all YAML form submissions.
   *
   * @param \Drupal\yamlform\YamlFormInterface|NULL $yamlform
   *   (optional) The YAML form to delete the submissions from.
   * @param int $limit
   *   (optional) Number of submissions to be deleted.
   * @param int $max_sid
   *   (optional) Maximum YAML form submission id.
   *
   * @return int
   *   The number of YAML form submissions deleted.
   */
  public function deleteAll(YamlFormInterface $yamlform = NULL, $limit = NULL, $max_sid = NULL) {
    $query = $this->getQuery()
      ->sort('sid');

    if ($yamlform) {
      $query->condition('yamlform_id', $yamlform->id());
    }
    if ($limit) {
      $query->range(0, $limit);
    }
    if ($max_sid) {
      $query->condition('sid', $max_sid, '<=');
    }

    $entity_ids = $query->execute();
    $entities = $this->loadMultiple($entity_ids);
    $this->delete($entities);
    return count($entities);
  }

  /**
   * Get the total number of submissions.
   *
   * @param \Drupal\yamlform\YamlFormInterface|NULL $yamlform
   *   (optional) A YAML form. If set the total number of submissions for the
   *   YAML form will be returned.
   * @param \Drupal\Core\Session\AccountInterface|NULL $account
   *   (optional) A user account.
   *
   * @return int
   *   Total number of submissions
   */
  public function getTotal(YamlFormInterface $yamlform = NULL, AccountInterface $account = NULL) {
    $query = $this->getQuery()->count();
    if ($yamlform) {
      $query->condition('yamlform_id', $yamlform->id());
    }
    if ($account) {
      $query->condition('uid', $account->id());
    }
    return $query->execute();
  }

  /**
   * Get the maximum sid.
   *
   * @param \Drupal\yamlform\YamlFormInterface|NULL $yamlform
   *   (optional) A YAML form. If set the total number of submissions for the
   *   YAML form will be returned.
   * @param \Drupal\Core\Session\AccountInterface|NULL $account
   *   (optional) A user account.
   *
   * @return int
   *   Total number of submissions
   */
  public function getMaxSubmissionId(YamlFormInterface $yamlform = NULL, AccountInterface $account = NULL) {
    $query = $this->getQuery();
    $query->sort('sid', 'DESC');
    if ($yamlform) {
      $query->condition('yamlform_id', $yamlform->id());
    }
    if ($account) {
      $query->condition('uid', $account->id());
    }
    $query->range(0, 1);
    return reset($query->execute());
  }

  /**
   * Get a YAML form submission's previous sibling.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user account.
   *
   * @return \Drupal\yamlform\YamlFormSubmissionInterface|NULL
   *   The YAML form submission's previous sibling.
   */
  public function getPreviousSubmission(YamlFormSubmissionInterface $yamlform_submission, AccountInterface $account) {
    return $this->getSiblingSubmission($yamlform_submission, $account, 'previous');
  }

  /**
   * Get a YAML form submission's next sibling.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user account.
   *
   * @return \Drupal\yamlform\YamlFormSubmissionInterface|NULL
   *   The YAML form submission's next sibling.
   */
  public function getNextSubmission(YamlFormSubmissionInterface $yamlform_submission, AccountInterface $account) {
    return $this->getSiblingSubmission($yamlform_submission, $account, 'next');
  }

  /**
   * Get a YAML form submission's previous or next sibling.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user account.
   * @param string $direction
   *   The direction, previous or next, of the sibling to return.
   *
   * @return \Drupal\yamlform\YamlFormSubmissionInterface|null
   *   The YAML form submission's previous or next sibling.
   */
  protected function getSiblingSubmission(YamlFormSubmissionInterface $yamlform_submission, AccountInterface $account, $direction = 'previous') {
    $yamlform = $yamlform_submission->getYamlForm();

    $query = $this->getQuery();
    $query->condition('yamlform_id', $yamlform->id());
    $query->range(0, 1);

    if (!$yamlform->access('view_any', $account)) {
      $query->condition('uid', $account->id());
    }

    if ($direction == 'previous') {
      $query->condition('sid', $yamlform_submission->id(), '<');
      $query->sort('sid', 'DESC');
    }
    else {
      $query->condition('sid', $yamlform_submission->id(), '>');
      $query->sort('sid', 'ASC');
    }

    return ($entity_ids = $query->execute()) ? $this->load(reset($entity_ids)) : NULL;
  }

}
