<?php

/**
 * @file
 * Contains \Drupal\yamlform\YamlFormSubmissionViewBuilder.
 */

namespace Drupal\yamlform;

use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityManagerInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityViewBuilder;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Utility\Token;
use Drupal\yamlform\Plugin\YamlFormElement\ContainerBase;
use Drupal\yamlform\Utility\YamlFormTidy;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Render controller for YAML form submissions.
 */
class YamlFormSubmissionViewBuilder extends EntityViewBuilder {

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The token handler.
   *
   * @var \Drupal\Core\Utility\Token $token
   */
  protected $token;

  /**
   * The YAML form handler manager service.
   *
   * @var \Drupal\yamlform\YamlFormHandlerManager
   */
  protected $yamlFormHandlerManager;

  /**
   * The YAML form element manager service.
   *
   * @var \Drupal\yamlform\YamlFormElementManager
   */
  protected $yamlFormElementManager;

  /**
   * Constructs a new YamlFormSubmissionViewBuilder.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type definition.
   * @param \Drupal\Core\Entity\EntityManagerInterface $entity_manager
   *   The entity manager service.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   * @param \Drupal\Core\Utility\Token $token
   *   The token handler.
   * @param \Drupal\yamlform\YamlFormHandlerManager $yamlform_handler_manager
   *   The YAML form handler manager service.
   * @param \Drupal\yamlform\YamlFormElementManager $yamlform_element_manager
   *   The YAML form element manager service.
   */
  public function __construct(EntityTypeInterface $entity_type, EntityManagerInterface $entity_manager, LanguageManagerInterface $language_manager, ConfigFactoryInterface $config_factory, Token $token, YamlFormHandlerManager $yamlform_handler_manager, YamlFormElementManager $yamlform_element_manager) {
    parent::__construct($entity_type, $entity_manager, $language_manager);
    $this->configFactory = $config_factory;
    $this->token = $token;
    $this->yamlFormHandlerManager = $yamlform_handler_manager;
    $this->yamlFormElementManager = $yamlform_element_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity.manager'),
      $container->get('language_manager'),
      $container->get('config.factory'),
      $container->get('token'),
      $container->get('plugin.manager.yamlform.handler'),
      $container->get('plugin.manager.yamlform.element')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildComponents(array &$build, array $entities, array $displays, $view_mode) {
    switch ($view_mode) {
      case 'yaml':
        foreach ($entities as $id => $entity) {
          $build[$id]['yaml'] = $this->buildYaml($entity);
        }
        break;

      case 'text':
        foreach ($entities as $id => $entity) {
          $build[$id]['text'] = $this->buildText($entity);
        }
        break;

      default:
        parent::buildComponents($build, $entities, $displays, $view_mode);
        foreach ($entities as $id => $entity) {
          // Link the entity id to the source entity.
          unset($build[$id]['entity_type']);
          if ($source_entity = $entity->getSourceEntity()) {
            $build[$id]['entity_id']['#title'] = 'Submitted to';
            $build[$id]['entity_id'][0] = [
              '#type' => 'link',
              '#title' => $source_entity->label(),
              '#url' => $source_entity->toUrl(),
            ];
          }

          // Link the source uri.
          if ($url = $entity->getSourceUrl()) {
            $build[$id]['uri'][0] = [
              '#type' => 'link',
              '#title' => $entity->uri->value,
              '#url' => $url,
            ];
          }

          // Convert YAML form submission to form items.
          unset($build[$id]['data']);
          if ($view_mode != 'summary') {
            $yamlform = $entity->getYamlForm();
            $data = $entity->getData();
            $inputs = $yamlform->getInputs();
            $build[$id]['data'] = $this->buildInputsAsItems($inputs, $data);
          }
        }
        break;

    }
  }

  /**
   * {@inheritdoc}
   */
  protected function alterBuild(array &$build, EntityInterface $entity, EntityViewDisplayInterface $display, $view_mode) {
    /* @var $entity \Drupal\yamlform\YamlFormSubmissionInterface */
    switch ($view_mode) {
      case 'yaml':
      case 'text':
        break;

      default:
        // Wrap submission information in a details widget.
        $build['submission'] = [
          '#type' => 'details',
          '#title' => $this->t('Submission information'),
          '#open' => TRUE,
          '#weight' => -10,
        ];
        foreach ($build as $key => $element) {
          if (!Element::property($key) && !in_array($key, ['data', 'submission'])) {
            $build['submission'][$key] = $element;
            // Only allow the Submission ID and Created date to be view by
            // everyone.
            if (!in_array($key, ['sid', 'created', 'entity_id'])) {
              $build['submission'][$key]['#access'] = $entity->getYamlForm()->access('submission_view_any');
            }
            unset($build[$key]);
          }
        }

        // Add navigation to full view.
        if ($view_mode == 'full') {
          $build['navigation'] = [
            '#theme' => 'yamlform_submission_navigation',
            '#yamlform_submission' => $entity,
            '#weight' => -20,
          ];
        }

        // Attach library.
        $build['#attached']['library'][] = 'yamlform/yamlform';
        break;
    }
  }

  /**
   * Build a YAML representation for a YAML form submission.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   *
   * @return array
   *   A render array displaying a YAML form submission as YAML.
   */
  protected function buildYaml(YamlFormSubmissionInterface $yamlform_submission) {
    $build = [];
    foreach ($yamlform_submission->getFieldDefinitions() as $key => $definition) {
      $item = $yamlform_submission->get($key);
      $value = (isset($item->target_id)) ? $item->target_id : $item->value;
      if ($key == 'data') {
        $value = str_replace("\n", "\n  ", YamlFormTidy::tidy($value));
        $build[$key] = [
          '#markup' => "$key: |\n  $value",
          '#suffix' => "\n",
        ];
      }
      else {
        if (in_array($key, ['remote_addr'])) {
          $value = "'$value'";
        }
        $build[$key] = [
          '#markup' => "$key: $value",
          '#suffix' => "\n",
        ];
      }
    }
    return $build;
  }

  /**
   * Build a plain text representation for a YAML form submission.
   *
   * @param \Drupal\yamlform\YamlFormSubmissionInterface $yamlform_submission
   *   A YAML form submission.
   *
   * @return array
   *   A render array displaying a YAML form submission as plain text.
   */
  protected function buildText(YamlFormSubmissionInterface $yamlform_submission) {
    $text = $this->configFactory->get('yamlform.settings')->get('display.submission_text');
    $token_data = [
      'yamlform' => $yamlform_submission->getYamlForm(),
      'yamlform-submission' => $yamlform_submission,
    ];
    return [
      '#markup' => $this->token->replace($text, $token_data),
    ];
  }

  /**
   * Build element display items from inputs and submitted data.
   *
   * @param array $elements
   *   A render array of form elements.
   * @param array $data
   *   Submission data.
   * @param array $options
   *   - excluded_inputs: An array of inputs to be excluded.
   * @param string $format
   *   Output format set to Html or Text.
   *
   * @return array
   *   A render array displaying the submitted values.
   */
  protected function buildInputsAsItems(array $elements, array $data, array $options = [], $format = 'Html') {
    $build_method = 'build' . $format;
    $build = [];

    foreach ($elements as $key => $element) {
      if (!is_array($element) || Element::property($key) || !$this->isVisibleElement($element) || isset($options['excluded_inputs'][$key])) {
        continue;
      }

      $element += ['#type' => NULL];
      $yamlform_element = $this->yamlFormElementManager->createInstance($element['#type']);
      if ($yamlform_element instanceof ContainerBase) {
        if ($children = $this->buildInputsAsItems($element, $data, $options, $format)) {
          $build[$key] = $yamlform_element->$build_method($element, $children);
        }
      }
      elseif (isset($data[$key]) && $data[$key] !== '') {
        $build[$key] = $yamlform_element->$build_method($element, $data[$key]);
      }
    }
    return $build;
  }

  /**
   * Build a very simple plain-text render of a submissions values.
   *
   * @param array $elements
   *   The elements/inputs.
   * @param array $data
   *   A YAML form submission data.
   * @param array $options
   *   - excluded_inputs: An array of inputs to be excluded.
   *
   * @return array
   *   A render array containing just plain next.
   */
  public function buildInputAsText(array $elements, array $data, array $options = []) {
    return $this->buildInputsAsItems($elements, $data, $options, 'Text');
  }

  /**
   * Determines if an element is visible.
   *
   * Copied from: \Drupal\Core\Render\Element::isVisibleElement
   * but does not hide hidden or value elements.
   *
   * @param array $element
   *   The element to check for visibility.
   *
   * @return bool
   *   TRUE if the element is visible, otherwise FALSE.
   */
  public function isVisibleElement(array $element) {
    return (isset($element['#type']))
      && (!isset($element['#access']) || (($element['#access'] instanceof AccessResultInterface && $element['#access']->isAllowed()) || ($element['#access'] === TRUE)));
  }

}
